# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.scheduler.cli.utils import lister_task_type


@pytest.mark.parametrize(
    "lister_name,listing_type", [("foo", "full"), ("bar", "incremental")]
)
def test_lister_task_type(lister_name, listing_type):
    assert lister_task_type(lister_name, listing_type) == (
        f"list-{lister_name}-{listing_type}"
    )



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Dict, List

from click.testing import CliRunner, Result
from confluent_kafka import Producer
import pytest
import yaml

from swh.journal.serializers import value_to_kafka
from swh.scheduler import get_scheduler
from swh.scheduler.cli import cli
from swh.scheduler.tests.test_journal_client import VISIT_STATUSES_1


@pytest.fixture
def swh_scheduler_cfg(postgresql_scheduler, kafka_server):
    """Journal client configuration ready"""
    return {
        "scheduler": {
            "cls": "postgresql",
            "db": postgresql_scheduler.dsn,
        },
        "journal": {
            "brokers": [kafka_server],
            "group_id": "test-consume-visit-status",
        },
    }


def _write_configuration_path(config: Dict, tmp_path: str) -> str:
    config_path = os.path.join(str(tmp_path), "scheduler.yml")
    with open(config_path, "w") as f:
        f.write(yaml.dump(config))
    return config_path


@pytest.fixture
def swh_scheduler_cfg_path(swh_scheduler_cfg, tmp_path):
    """Write scheduler configuration in temporary path and returns such path"""
    return _write_configuration_path(swh_scheduler_cfg, tmp_path)


def invoke(args: List[str], config_path: str, catch_exceptions: bool = False) -> Result:
    """Invoke swh scheduler journal subcommands"""
    runner = CliRunner()
    result = runner.invoke(cli, ["-C" + config_path] + args)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test_cli_journal_client_origin_visit_status_misconfiguration_no_scheduler(
    swh_scheduler_cfg, tmp_path
):
    config = swh_scheduler_cfg.copy()
    config["scheduler"] = {"cls": "foo"}
    config_path = _write_configuration_path(config, tmp_path)
    with pytest.raises(ValueError, match="must be instantiated"):
        invoke(
            [
                "journal-client",
                "--stop-after-objects",
                "1",
            ],
            config_path,
        )


def test_cli_journal_client_origin_visit_status_misconfiguration_missing_journal_conf(
    swh_scheduler_cfg, tmp_path
):
    config = swh_scheduler_cfg.copy()
    config.pop("journal", None)
    config_path = _write_configuration_path(config, tmp_path)

    with pytest.raises(ValueError, match="Missing 'journal'"):
        invoke(
            [
                "journal-client",
                "--stop-after-objects",
                "1",
            ],
            config_path,
        )


def test_cli_journal_client_origin_visit_status(
    swh_scheduler_cfg,
    swh_scheduler_cfg_path,
):
    kafka_server = swh_scheduler_cfg["journal"]["brokers"][0]
    swh_scheduler = get_scheduler(**swh_scheduler_cfg["scheduler"])
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test visit-stats producer",
            "acks": "all",
        }
    )
    visit_status = VISIT_STATUSES_1[0]

    value = value_to_kafka(visit_status)
    topic = "swh.journal.objects.origin_visit_status"
    producer.produce(topic=topic, key=b"bogus-origin", value=value)
    producer.flush()

    result = invoke(
        [
            "journal-client",
            "--stop-after-objects",
            "1",
        ],
        swh_scheduler_cfg_path,
    )

    # Check the output
    expected_output = "Processed 1 message(s).\nDone.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    actual_visit_stats = swh_scheduler.origin_visit_stats_get(
        [(visit_status["origin"], visit_status["type"])]
    )

    assert actual_visit_stats
    assert len(actual_visit_stats) == 1



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Tuple

import pytest

from swh.scheduler.cli.origin import format_origins
from swh.scheduler.tests.common import TASK_TYPES
from swh.scheduler.tests.test_cli import invoke as basic_invoke


def invoke(scheduler, args: Tuple[str, ...] = (), catch_exceptions: bool = False):
    return basic_invoke(
        scheduler, args=["origin", *args], catch_exceptions=catch_exceptions
    )


def test_cli_origin(swh_scheduler):
    """Check that swh scheduler origin returns its help text"""

    result = invoke(swh_scheduler)

    assert "Commands:" in result.stdout


def test_format_origins_basic(listed_origins):
    listed_origins = listed_origins[:100]

    basic_output = list(format_origins(listed_origins))
    # 1 header line + all origins
    assert len(basic_output) == len(listed_origins) + 1

    no_header_output = list(format_origins(listed_origins, with_header=False))
    assert basic_output[1:] == no_header_output


def test_format_origins_fields_unknown(listed_origins):
    listed_origins = listed_origins[:10]

    it = format_origins(listed_origins, fields=["unknown_field"])

    with pytest.raises(ValueError, match="unknown_field"):
        next(it)


def test_format_origins_fields(listed_origins):
    listed_origins = listed_origins[:10]
    fields = ["lister_id", "url", "visit_type"]

    output = list(format_origins(listed_origins, fields=fields))
    assert output[0] == ",".join(fields)
    for i, origin in enumerate(listed_origins):
        assert output[i + 1] == f"{origin.lister_id},{origin.url},{origin.visit_type}"


def test_grab_next(swh_scheduler, listed_origins_by_type):
    NUM_RESULTS = 10
    # Strict inequality to check that grab_next_visits doesn't return more
    # results than requested

    # XXX: should test all of 'listed_origins_by_type' here...
    visit_type = next(iter(listed_origins_by_type))
    assert len(listed_origins_by_type[visit_type]) > NUM_RESULTS

    for origins in listed_origins_by_type.values():
        swh_scheduler.record_listed_origins(origins)

    result = invoke(swh_scheduler, args=("grab-next", visit_type, str(NUM_RESULTS)))
    assert result.exit_code == 0

    out_lines = result.stdout.splitlines()
    assert len(out_lines) == NUM_RESULTS + 1

    fields = out_lines[0].split(",")
    returned_origins = [dict(zip(fields, line.split(","))) for line in out_lines[1:]]

    # Check that we've received origins we had listed in the first place
    assert set(origin["url"] for origin in returned_origins) <= set(
        origin.url for origin in listed_origins_by_type[visit_type]
    )


def test_schedule_next(swh_scheduler, listed_origins_by_type):
    for task_type in TASK_TYPES.values():
        swh_scheduler.create_task_type(task_type)

    NUM_RESULTS = 10
    # Strict inequality to check that grab_next_visits doesn't return more
    # results than requested
    visit_type = next(iter(listed_origins_by_type))
    assert len(listed_origins_by_type[visit_type]) > NUM_RESULTS

    for origins in listed_origins_by_type.values():
        swh_scheduler.record_listed_origins(origins)

    result = invoke(swh_scheduler, args=("schedule-next", visit_type, str(NUM_RESULTS)))
    assert result.exit_code == 0

    # pull all tasks out of the scheduler
    tasks = swh_scheduler.search_tasks()
    assert len(tasks) == NUM_RESULTS

    scheduled_tasks = {
        (task["type"], task["arguments"]["kwargs"]["url"]) for task in tasks
    }
    all_possible_tasks = {
        (f"load-{origin.visit_type}", origin.url)
        for origin in listed_origins_by_type[visit_type]
    }

    assert scheduled_tasks <= all_possible_tasks


def test_send_to_celery_unknown_visit_type(
    swh_scheduler,
):
    "Calling cli without a known visit type should raise"
    with pytest.raises(ValueError, match="Unknown"):
        invoke(swh_scheduler, args=("send-to-celery", "unknown-visit-type"))


@pytest.mark.parametrize(
    "extra_cmd_args",
    [[], ["--lister-name", "github", "--lister-instance-name", "github"]],
)
def test_send_to_celery(
    mocker,
    swh_scheduler,
    swh_scheduler_celery_app,
    listed_origins_by_type,
    extra_cmd_args,
):
    for task_type in TASK_TYPES.values():
        swh_scheduler.create_task_type(task_type)

    visit_type = next(iter(listed_origins_by_type))

    for origins in listed_origins_by_type.values():
        swh_scheduler.record_listed_origins(origins)

    get_queue_length = mocker.patch(
        "swh.scheduler.celery_backend.config.get_queue_length"
    )
    get_queue_length.return_value = None

    send_task = mocker.patch.object(swh_scheduler_celery_app, "send_task")
    send_task.return_value = None

    cmd_args = ["send-to-celery", visit_type] + extra_cmd_args

    result = invoke(swh_scheduler, args=tuple(cmd_args))
    assert result.exit_code == 0

    scheduled_tasks = {
        (call[0][0], call[1]["kwargs"]["url"]) for call in send_task.call_args_list
    }

    expected_tasks = {
        (TASK_TYPES[origin.visit_type]["backend_name"], origin.url)
        for origin in listed_origins_by_type[visit_type]
    }

    assert expected_tasks == scheduled_tasks


def test_update_metrics(swh_scheduler, listed_origins):
    swh_scheduler.record_listed_origins(listed_origins)

    assert swh_scheduler.get_metrics() == []

    result = invoke(swh_scheduler, args=("update-metrics",))

    assert result.exit_code == 0
    assert swh_scheduler.get_metrics() != []



# Copyright (C) 2021-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""This schedules the recurrent visits, for listed origins, in Celery.

For "oneshot" (save code now, lister) tasks, check the
:mod:`swh.scheduler.celery_backend.runner` and
:mod:`swh.scheduler.celery_backend.pika_listener` modules.

"""

from __future__ import annotations

from itertools import chain
import logging
from queue import Empty, Queue
import random
from threading import Thread
import time
from typing import TYPE_CHECKING, Any, Dict, List, Tuple

from kombu.utils.uuid import uuid

from swh.scheduler.celery_backend.config import get_available_slots
from swh.scheduler.utils import create_origin_task_dicts

if TYPE_CHECKING:
    from ..interface import SchedulerInterface
    from ..model import ListedOrigin

logger = logging.getLogger(__name__)


DEFAULT_POLICY = [
    {"policy": "already_visited_order_by_lag", "weight": 50},
    {"policy": "never_visited_oldest_update_first", "weight": 50},
]
DEFAULT_DVCS_POLICY = [
    {"policy": "already_visited_order_by_lag", "weight": 49},
    {"policy": "never_visited_oldest_update_first", "weight": 49},
    {"policy": "origins_without_last_update", "weight": 2},
]
DEFAULT_GIT_POLICY = [
    {"policy": "already_visited_order_by_lag", "weight": 49, "tablesample": 0.1},
    {"policy": "never_visited_oldest_update_first", "weight": 49, "tablesample": 0.1},
    {"policy": "origins_without_last_update", "weight": 2, "tablesample": 0.1},
]

DEFAULT_POLICY_CONFIG: Dict[str, List[Dict[str, Any]]] = {
    "default": DEFAULT_POLICY,
    "hg": DEFAULT_DVCS_POLICY,
    "svn": DEFAULT_DVCS_POLICY,
    "cvs": DEFAULT_DVCS_POLICY,
    "bzr": DEFAULT_DVCS_POLICY,
    "git": DEFAULT_GIT_POLICY,
}

"""Scheduling policies to use to retrieve visits for the given visit types, with their
relative weights"""

MIN_SLOTS_RATIO = 0.05
"""Quantity of slots that need to be available (with respect to max_queue_length) for
:py:func:`~swh.scheduler.interface.SchedulerInterface.grab_next_visits` to trigger"""

QUEUE_FULL_BACKOFF = 60
"""Backoff time (in seconds) if there's fewer than :py:data:`MIN_SLOTS_RATIO` slots
available in the queue."""

DEFAULT_NO_ORIGINS_SCHEDULED_BACKOFF = 20 * 60
"""Backoff time (in seconds) if no origins have been scheduled in the current
iteration"""

BACKOFF_SPLAY = 5.0
"""Amplitude of the fuzziness between backoffs"""

TERMINATE = object()
"""Termination request received from command queue (singleton used for identity
comparison)"""


def grab_next_visits_policy_weights(
    scheduler: SchedulerInterface,
    visit_type: str,
    num_visits: int,
    policy_cfg: List[Dict[str, Any]],
) -> List[ListedOrigin]:
    """Get the next ``num_visits`` for the given ``visit_type`` using the corresponding
    set of scheduling policies.

    The :py:data:`POLICY_CFG` list sets, for the current visit type, the
    scheduling policies used to pull the next tasks. Each policy config entry
    in the list should at least have a 'policy' (policy name) and a 'weight'
    entry. For each policy in this policy_cfg list, the number of returned
    origins to visit will be weighted using this weight config option so that
    the total number of returned origins is around num_visits. Any other
    key/value entry in the policy configuration will be passed to the
    scheduler.grab_next_visit() method.

    This function emits a warning if the ratio of retrieved origins is off of
    the requested ratio by more than 5%.

    Returns:
       at most ``num_visits`` :py:class:`~swh.scheduler.model.ListedOrigin` objects

    """

    policies = [cfg["policy"] for cfg in policy_cfg]
    if len(set(policies)) != len(policies):
        raise ValueError(
            "A policy weights can only appear once; check your policy "
            f"configuration for visit type {visit_type}"
        )

    weights = [cfg["weight"] for cfg in policy_cfg]
    total_weight = sum(weights)
    if not total_weight:
        raise ValueError(f"No policy weights set for visit type {visit_type}")

    ratios = [weight / total_weight for weight in weights]

    extra_kws = [
        {k: v for k, v in cfg.items() if k not in ("weight", "policy")}
        for cfg in policy_cfg
    ]

    fetched_origins: Dict[str, List[ListedOrigin]] = {}
    for policy, ratio, extra_kw in zip(policies, ratios, extra_kws):
        num_tasks_to_send = int(num_visits * ratio)
        fetched_origins[policy] = scheduler.grab_next_visits(
            visit_type,
            num_tasks_to_send,
            policy=policy,
            **extra_kw,
        )

    all_origins: List[ListedOrigin] = list(
        chain.from_iterable(fetched_origins.values())
    )
    if not all_origins:
        return []

    # Check whether the ratios of origins fetched are skewed with respect to the
    # ones we requested
    fetched_origin_ratios = {
        policy: len(origins) / len(all_origins)
        for policy, origins in fetched_origins.items()
    }

    for policy, expected_ratio in zip(policies, ratios):
        # 5% of skew with respect to request
        if abs(fetched_origin_ratios[policy] - expected_ratio) / expected_ratio > 0.05:
            logger.info(
                "Skewed fetch for visit type %s with policy %s: fetched %s, "
                "requested %s",
                visit_type,
                policy,
                fetched_origin_ratios[policy],
                expected_ratio,
            )

    return all_origins


def splay():
    """Return a random short interval by which to vary the backoffs for the visit
    scheduling threads"""
    return random.uniform(0, BACKOFF_SPLAY)


def send_visits_for_visit_type(
    scheduler: SchedulerInterface,
    app,
    visit_type: str,
    task_type: Dict,
    policy_cfg: List[Dict[str, Any]],
    no_origins_scheduled_backoff: int = DEFAULT_NO_ORIGINS_SCHEDULED_BACKOFF,
) -> float:
    """Schedule the next batch of visits for the given ``visit_type``.

    First, we determine the number of available slots by introspecting the RabbitMQ
    queue.

    If there's fewer than :py:data:`MIN_SLOTS_RATIO` slots available in the queue, we
    wait for :py:data:`QUEUE_FULL_BACKOFF` seconds. This avoids running the expensive
    :py:func:`~swh.scheduler.interface.SchedulerInterface.grab_next_visits` queries when
    there's not many jobs to queue.

    Once there's more than :py:data:`MIN_SLOTS_RATIO` slots available, we run
    :py:func:`grab_next_visits_policy_weights` to retrieve the next set of origin visits
    to schedule, and we send them to celery.

    If the last scheduling attempt didn't return any origins, we sleep by default for
    :py:data:`DEFAULT_NO_ORIGINS_SCHEDULED_BACKOFF` seconds. This avoids running the expensive
    :py:func:`~swh.scheduler.interface.SchedulerInterface.grab_next_visits` queries too
    often if there's nothing left to schedule.

    The :py:data:`POLICY_CFG` argument is the policy configuration used to
    choose the next origins to visit. It is passed directly to the
    :py:func:`grab_next_visits_policy_weights()` function.

    Returns:
       the earliest :py:func:`time.monotonic` value at which to run the next iteration
       of the loop.

    """
    queue_name = task_type["backend_name"]
    max_queue_length = task_type.get("max_queue_length") or 0
    min_available_slots = max_queue_length * MIN_SLOTS_RATIO

    current_iteration_start = time.monotonic()

    # Check queue level
    available_slots = get_available_slots(app, queue_name, max_queue_length)
    logger.debug(
        "%s available slots for visit type %s in queue %s",
        available_slots,
        visit_type,
        queue_name,
    )
    if available_slots < min_available_slots:
        return current_iteration_start + QUEUE_FULL_BACKOFF

    origins = grab_next_visits_policy_weights(
        scheduler, visit_type, available_slots, policy_cfg
    )

    if not origins:
        logger.debug("No origins to visit for type %s", visit_type)
        return current_iteration_start + no_origins_scheduled_backoff

    # Try to smooth the ingestion load, origins pulled by different
    # scheduling policies have different resource usage patterns
    random.shuffle(origins)

    for task_dict in create_origin_task_dicts(origins, scheduler):
        app.send_task(
            queue_name,
            task_id=uuid(),
            args=task_dict["arguments"]["args"],
            kwargs=task_dict["arguments"]["kwargs"],
            queue=queue_name,
        )

    logger.info(
        "%s: %s visits scheduled in queue %s",
        visit_type,
        len(origins),
        queue_name,
    )

    # When everything worked, we can try to schedule origins again ASAP.
    return time.monotonic()


def visit_scheduler_thread(
    config: Dict,
    visit_type: str,
    command_queue: Queue[object],
    exc_queue: Queue[Tuple[str, BaseException]],
):
    """Target function for the visit sending thread, which initializes local connections
    and handles exceptions by sending them back to the main thread."""

    from swh.scheduler import get_scheduler
    from swh.scheduler.celery_backend.config import build_app

    try:
        # We need to reinitialize these connections because they're not generally
        # thread-safe
        app = build_app(config.get("celery"))
        scheduler = get_scheduler(**config["scheduler"])
        task_name = f"load-{visit_type}"
        task_type = scheduler.get_task_type(task_name)
        if task_type is None:
            raise ValueError(f"Unknown task type: {task_name}")

        policy_cfg = config.get("scheduling_policy", DEFAULT_POLICY_CONFIG)
        for policies in policy_cfg.values():
            for policy in policies:
                if "weight" not in policy or "policy" not in policy:
                    raise ValueError(
                        "Each policy configuration needs at least a 'policy' "
                        "and a 'weight' entry"
                    )
        policy_cfg = {**DEFAULT_POLICY_CONFIG, **policy_cfg}

        next_iteration = time.monotonic()

        while True:
            # vary the next iteration time a little bit
            next_iteration = next_iteration + splay()
            while time.monotonic() < next_iteration:
                # Wait for next iteration to start. Listen for termination message.
                try:
                    msg = command_queue.get(block=True, timeout=1)
                except Empty:
                    continue

                if msg is TERMINATE:
                    return
                else:
                    logger.warn("Received unexpected message %s in command queue", msg)

            # Refresh the task_type object from the database for new parameters, e.g.
            # the max queue length
            task_type = scheduler.get_task_type(task_name)
            if task_type is None:
                raise ValueError(f"Unknown task type: {task_name}")

            next_iteration = send_visits_for_visit_type(
                scheduler,
                app,
                visit_type,
                task_type,
                policy_cfg.get(visit_type, policy_cfg["default"]),
                config.get(
                    "no_origins_scheduled_backoff", DEFAULT_NO_ORIGINS_SCHEDULED_BACKOFF
                ),
            )

    except BaseException as e:
        exc_queue.put((visit_type, e))


VisitSchedulerThreads = Dict[str, Tuple[Thread, Queue]]
"""Dict storing the visit scheduler threads and their command queues"""


def spawn_visit_scheduler_thread(
    threads: VisitSchedulerThreads,
    exc_queue: Queue[Tuple[str, BaseException]],
    config: Dict[str, Any],
    visit_type: str,
):
    """Spawn a new thread to schedule the visits of type ``visit_type``."""
    command_queue: Queue[object] = Queue()
    thread = Thread(
        target=visit_scheduler_thread,
        kwargs={
            "config": config,
            "visit_type": visit_type,
            "command_queue": command_queue,
            "exc_queue": exc_queue,
        },
    )
    threads[visit_type] = (thread, command_queue)
    thread.start()


def terminate_visit_scheduler_threads(threads: VisitSchedulerThreads) -> List[str]:
    """Terminate all visit scheduler threads"""
    logger.info("Termination requested...")
    for _, command_queue in threads.values():
        command_queue.put(TERMINATE)

    loops = 0
    while threads and loops < 10:
        logger.info(
            "Terminating visit scheduling threads: %s", ", ".join(sorted(threads))
        )
        loops += 1
        for visit_type, (thread, _) in list(threads.items()):
            thread.join(timeout=1)
            if not thread.is_alive():
                logger.debug("Thread %s terminated", visit_type)
                del threads[visit_type]

    if threads:
        logger.warn(
            "Could not reap the following threads after 10 attempts: %s",
            ", ".join(sorted(threads)),
        )

    return list(sorted(threads))



# Copyright (C) 2019-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control

from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Tuple

    from swh.scheduler.interface import SchedulerInterface


TASK_BATCH_SIZE = 1000  # Number of tasks per query to the scheduler


def schedule_origin_batches(scheduler, task_type, origins, origin_batch_size, kwargs):
    from itertools import islice

    from swh.scheduler.utils import create_task_dict

    nb_origins = 0
    nb_tasks = 0

    while True:
        task_batch = []
        for _ in range(TASK_BATCH_SIZE):
            # Group origins
            origin_batch = []
            for origin in islice(origins, origin_batch_size):
                origin_batch.append(origin)
            nb_origins += len(origin_batch)
            if not origin_batch:
                break

            # Create a task for these origins
            args = [origin_batch]
            task_dict = create_task_dict(task_type, "oneshot", *args, **kwargs)
            task_batch.append(task_dict)

        # Schedule a batch of tasks
        if not task_batch:
            break
        nb_tasks += len(task_batch)
        if scheduler:
            scheduler.create_tasks(task_batch)
        click.echo("Scheduled %d tasks (%d origins)." % (nb_tasks, nb_origins))

    # Print final status.
    if nb_tasks:
        click.echo("Done.")
    else:
        click.echo("Nothing to do (no origin metadata matched the criteria).")


def parse_argument(option):
    import yaml

    if option == "":
        # yaml.safe_load("") returns None
        return ""

    try:
        return yaml.safe_load(option)
    except Exception:
        raise click.ClickException("Invalid argument: {}".format(option))


def parse_options(options: List[str]) -> Tuple[List[str], Dict]:
    """Parses options from a CLI as YAML and turns it into Python
    args and kwargs.

    >>> parse_options([])
    ([], {})
    >>> parse_options(['foo', 'bar'])
    (['foo', 'bar'], {})
    >>> parse_options(['[foo, bar]'])
    ([['foo', 'bar']], {})
    >>> parse_options(['"foo"', '"bar"'])
    (['foo', 'bar'], {})
    >>> parse_options(['foo="bar"'])
    ([], {'foo': 'bar'})
    >>> parse_options(['"foo"', 'bar="baz"'])
    (['foo'], {'bar': 'baz'})
    >>> parse_options(['42', 'bar=False'])
    ([42], {'bar': False})
    >>> parse_options(['42', 'bar=false'])
    ([42], {'bar': False})
    >>> parse_options(['foo', ''])
    (['foo', ''], {})
    >>> parse_options(['foo', 'bar='])
    (['foo'], {'bar': ''})
    >>> parse_options(['foo', 'null'])
    (['foo', None], {})
    >>> parse_options(['foo', 'bar=null'])
    (['foo'], {'bar': None})
    >>> parse_options(['42', '"foo'])
    Traceback (most recent call last):
      ...
    click.exceptions.ClickException: Invalid argument: "foo
    """
    kw_pairs = [x.split("=", 1) for x in options if "=" in x]
    args = [parse_argument(x) for x in options if "=" not in x]
    kw = {k: parse_argument(v) for (k, v) in kw_pairs}
    return (args, kw)


def get_task_type(scheduler: SchedulerInterface, visit_type: str) -> Optional[Dict]:
    "Given a visit type, return its associated task type."
    return scheduler.get_task_type(f"load-{visit_type}")


def send_to_celery(
    scheduler: SchedulerInterface,
    visit_type_to_queue: Dict[str, str],
    enabled: bool = True,
    lister_name: Optional[str] = None,
    lister_instance_name: Optional[str] = None,
    policy: str = "oldest_scheduled_first",
    tablesample: Optional[float] = None,
):
    """Utility function to read tasks from the scheduler and send those directly to
    celery.

    Args:
        visit_type_to_queue: Optional mapping of visit/loader type (e.g git, svn, ...)
          to queue to send task to.
        enabled: Determine whether we want to list enabled or disabled origins. As
          default, we want reasonably enabled origins. For some edge case, we might
          want the others.
        lister_name: Determine the list of origins listed from the lister with name
        lister_instance_name: Determine the list of origins listed from the lister
          with instance name
        policy: the scheduling policy used to select which visits to schedule
        tablesample: the percentage of the table on which we run the query
          (None: no sampling)

    """

    from kombu.utils.uuid import uuid

    from swh.scheduler.celery_backend.config import app, get_available_slots

    from ..utils import create_origin_task_dicts

    for visit_type_name, queue_name in visit_type_to_queue.items():
        task_type = get_task_type(scheduler, visit_type_name)
        assert task_type is not None
        task_name = task_type["backend_name"]
        num_tasks = get_available_slots(app, queue_name, task_type["max_queue_length"])

        click.echo(f"{num_tasks} slots available in celery queue")

        origins = scheduler.grab_next_visits(
            visit_type_name,
            num_tasks,
            policy=policy,
            tablesample=tablesample,
            enabled=enabled,
            lister_name=lister_name,
            lister_instance_name=lister_instance_name,
        )

        click.echo(f"{len(origins)} visits to send to celery")
        for task_dict in create_origin_task_dicts(origins, scheduler):
            app.send_task(
                task_name,
                task_id=uuid(),
                args=task_dict["arguments"]["args"],
                kwargs=task_dict["arguments"]["kwargs"],
                queue=queue_name,
            )


def pretty_print_list(list, indent=0):
    """Pretty-print a list"""
    return "".join("%s%r\n" % (" " * indent, item) for item in list)


def pretty_print_dict(dict, indent=0):
    """Pretty-print a list"""
    return "".join(
        "%s%s: %r\n" % (" " * indent, click.style(key, bold=True), value)
        for key, value in sorted(dict.items())
    )


def format_dict(d):
    """Recursively format date objects in the dict passed as argument"""
    import datetime

    ret = {}
    for k, v in d.items():
        if isinstance(v, (datetime.date, datetime.datetime)):
            v = v.isoformat()
        elif isinstance(v, dict):
            v = format_dict(v)
        ret[k] = v
    return ret


def pretty_print_run(run, indent=4):
    fmt = (
        "{indent}{backend_id} [{status}]\n"
        "{indent}  scheduled: {scheduled} [{started}:{ended}]"
    )
    return fmt.format(indent=" " * indent, **format_dict(run))


def pretty_print_task(task, full=False):
    """Pretty-print a task

    If 'full' is True, also print the status and priority fields.

    >>> import datetime
    >>> task = {
    ...     'id': 1234,
    ...     'arguments': {
    ...         'args': ['foo', 'bar', True],
    ...         'kwargs': {'key': 'value', 'key2': 42},
    ...     },
    ...     'current_interval': datetime.timedelta(hours=1),
    ...     'next_run': datetime.datetime(2019, 2, 21, 13, 52, 35, 407818),
    ...     'policy': 'oneshot',
    ...     'priority': None,
    ...     'status': 'next_run_not_scheduled',
    ...     'type': 'test_task',
    ... }
    >>> print(click.unstyle(pretty_print_task(task)))
    Task 1234
      Next run: ... (2019-02-21T13:52:35.407818)
      Interval: 1:00:00
      Type: test_task
      Policy: oneshot
      Args:
        'foo'
        'bar'
        True
      Keyword args:
        key: 'value'
        key2: 42
    <BLANKLINE>
    >>> print(click.unstyle(pretty_print_task(task, full=True)))
    Task 1234
      Next run: ... (2019-02-21T13:52:35.407818)
      Interval: 1:00:00
      Type: test_task
      Policy: oneshot
      Status: next_run_not_scheduled
      Priority:\x20
      Args:
        'foo'
        'bar'
        True
      Keyword args:
        key: 'value'
        key2: 42
    <BLANKLINE>
    """
    import humanize

    next_run = task["next_run"]
    lines = [
        "%s %s\n" % (click.style("Task", bold=True), task["id"]),
        click.style("  Next run: ", bold=True),
        "%s (%s)" % (humanize.naturaldate(next_run), next_run.isoformat()),
        "\n",
        click.style("  Interval: ", bold=True),
        str(task["current_interval"]),
        "\n",
        click.style("  Type: ", bold=True),
        task["type"] or "",
        "\n",
        click.style("  Policy: ", bold=True),
        task["policy"] or "",
        "\n",
    ]
    if full:
        lines += [
            click.style("  Status: ", bold=True),
            task["status"] or "",
            "\n",
            click.style("  Priority: ", bold=True),
            task["priority"] or "",
            "\n",
        ]
    lines += [
        click.style("  Args:\n", bold=True),
        pretty_print_list(task["arguments"]["args"], indent=4),
        click.style("  Keyword args:\n", bold=True),
        pretty_print_dict(task["arguments"]["kwargs"], indent=4),
    ]

    return "".join(lines)


def task_add(
    scheduler: SchedulerInterface,
    task_type_name: str,
    args: List[str],
    kw: Dict,
    policy: str,
    priority: Optional[str] = None,
    next_run: Optional[str] = None,
):
    """Add a task task_type_name in the scheduler."""
    from swh.scheduler.utils import utcnow

    task = {
        "type": task_type_name,
        "policy": policy,
        "priority": priority,
        "arguments": {
            "args": args,
            "kwargs": kw,
        },
        "next_run": next_run or utcnow(),
    }
    created = scheduler.create_tasks([task])

    output = [f"Created {len(created)} tasks\n"]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo("\n".join(output))


def lister_task_type(lister_name: str, lister_type: Optional[str] = None) -> str:
    """Compute expected scheduler task type from the lister name and its optional
    listing type (full, incremental).

    """
    prefix = f"list-{lister_name}"
    return f"{prefix}-{lister_type}" if lister_type else prefix



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.scheduler.cli.utils import lister_task_type


@pytest.mark.parametrize(
    "lister_name, lister_type, expected_task_type",
    [
        ["cgit", None, "list-cgit"],
        ["gitlab", "full", "list-gitlab-full"],
        ["gitea", "incremental", "list-gitea-incremental"],
    ],
)
def test_lister_task_type(lister_name, lister_type, expected_task_type):
    assert lister_task_type(lister_name, lister_type) == expected_task_type



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import functools
from itertools import permutations
from typing import List
from unittest.mock import Mock

import attr
import pytest

from swh.model.hashutil import hash_to_bytes
from swh.scheduler.journal_client import (
    from_position_offset_to_days,
    max_date,
    next_visit_queue_position,
    process_journal_objects,
)
from swh.scheduler.model import LastVisitStatus, ListedOrigin, OriginVisitStats
from swh.scheduler.utils import utcnow


def test_journal_client_origin_visit_status_from_journal_fail(swh_scheduler):
    process_fn = functools.partial(
        process_journal_objects,
        scheduler=swh_scheduler,
    )

    with pytest.raises(AssertionError, match="Got unexpected origin_visit"):
        process_fn(
            {
                "origin_visit": [
                    {"url": "http://foobar.baz"},
                ]
            }
        )

    with pytest.raises(AssertionError, match="Expected origin_visit_status"):
        process_fn({})


ONE_DAY = datetime.timedelta(days=1)
ONE_YEAR = datetime.timedelta(days=366)

DATE3 = utcnow()
DATE2 = DATE3 - ONE_DAY
DATE1 = DATE2 - ONE_DAY


assert DATE1 < DATE2 < DATE3


@pytest.mark.parametrize(
    "dates,expected_max_date",
    [
        ((DATE1,), DATE1),
        ((None, DATE2), DATE2),
        ((DATE1, None), DATE1),
        ((DATE1, DATE2), DATE2),
        ((DATE2, DATE1), DATE2),
        ((DATE1, DATE2, DATE3), DATE3),
        ((None, DATE2, DATE3), DATE3),
        ((None, None, DATE3), DATE3),
        ((DATE1, None, DATE3), DATE3),
    ],
)
def test_max_date(dates, expected_max_date):
    assert max_date(*dates) == expected_max_date


def test_max_date_raise():
    with pytest.raises(ValueError, match="valid datetime"):
        max_date()
    with pytest.raises(ValueError, match="valid datetime"):
        max_date(None)
    with pytest.raises(ValueError, match="valid datetime"):
        max_date(None, None)


def test_journal_client_origin_visit_status_from_journal_ignored_status(swh_scheduler):
    """Only final statuses (full, partial) are important, the rest remain ignored."""
    # Trace method calls on the swh_scheduler
    swh_scheduler = Mock(wraps=swh_scheduler)

    visit_statuses = [
        {
            "origin": "foo",
            "visit": 1,
            "status": "created",
            "date": utcnow(),
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 1,
            "status": "ongoing",
            "date": utcnow(),
            "type": "svn",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    # All messages have been ignored: no stats have been upserted
    swh_scheduler.origin_visit_stats_upsert.assert_not_called()


def test_journal_client_ignore_missing_type(swh_scheduler):
    """Ignore statuses with missing type key"""
    # Trace method calls on the swh_scheduler
    swh_scheduler = Mock(wraps=swh_scheduler)

    date = utcnow()
    snapshot = hash_to_bytes("dddcc0710eb6cf9efd5b920a8453e1e07157bddd")
    visit_statuses = [
        {
            "origin": "foo",
            "visit": 1,
            "status": "full",
            "date": date,
            "snapshot": snapshot,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    # The message has been ignored: no stats have been upserted
    swh_scheduler.origin_visit_stats_upsert.assert_not_called()


def assert_visit_stats_ok(
    actual_visit_stats: OriginVisitStats,
    expected_visit_stats: OriginVisitStats,
    ignore_fields: List[str] = ["next_visit_queue_position"],
):
    """Utility test function to ensure visits stats read from the backend are in the right
    shape. The comparison on the next_visit_queue_position will be dealt with in
    dedicated tests so it's not tested in tests that are calling this function.

    """
    fields = attr.fields_dict(OriginVisitStats)
    defaults = {field: fields[field].default for field in ignore_fields}

    actual_visit_stats = attr.evolve(actual_visit_stats, **defaults)
    assert actual_visit_stats == expected_visit_stats


def test_journal_client_origin_visit_status_from_journal_last_not_found(swh_scheduler):
    visit_status = {
        "origin": "foo",
        "visit": 1,
        "status": "not_found",
        "date": DATE1,
        "type": "git",
        "snapshot": None,
    }

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_visit=visit_status["date"],
            last_visit_status=LastVisitStatus.not_found,
            next_position_offset=4,
            successive_visits=1,
        ),
    )

    visit_statuses = [
        {
            "origin": "foo",
            "visit": 3,
            "status": "not_found",
            "date": DATE2,
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "foo",
            "visit": 4,
            "status": "not_found",
            "date": DATE3,
            "type": "git",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.not_found,
            next_position_offset=6,
            successive_visits=3,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_failed(swh_scheduler):
    visit_statuses = [
        {
            "origin": "foo",
            "visit": 1,
            "status": "partial",
            "date": utcnow(),
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 1,
            "status": "full",
            "date": DATE1,
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 2,
            "status": "full",
            "date": DATE2,
            "type": "git",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "full",
            "date": DATE3,
            "type": "git",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("bar", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="bar",
            visit_type="git",
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.failed,
            next_position_offset=6,
            successive_visits=3,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_failed2(swh_scheduler):
    visit_statuses = [
        {
            "origin": "bar",
            "visit": 2,
            "status": "failed",
            "date": DATE1,
            "type": "git",
            "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "failed",
            "date": DATE2,
            "type": "git",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("bar", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="bar",
            visit_type="git",
            last_visit=DATE2,
            last_visit_status=LastVisitStatus.failed,
            next_position_offset=5,
            successive_visits=2,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_successful(swh_scheduler):
    visit_statuses = [
        {
            "origin": "bar",
            "visit": 1,
            "status": "partial",
            "date": utcnow(),
            "type": "git",
            "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        },
        {
            "origin": "foo",
            "visit": 1,
            "status": "full",
            "date": DATE1,
            "type": "git",
            "snapshot": hash_to_bytes("eeecc0710eb6cf9efd5b920a8453e1e07157bfff"),
        },
        {
            "origin": "foo",
            "visit": 2,
            "status": "partial",
            "date": DATE2,
            "type": "git",
            "snapshot": hash_to_bytes("aaacc0710eb6cf9efd5b920a8453e1e07157baaa"),
        },
        {
            "origin": "foo",
            "visit": 3,
            "status": "full",
            "date": DATE3,
            "type": "git",
            "snapshot": hash_to_bytes("dddcc0710eb6cf9efd5b920a8453e1e07157bddd"),
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE3,
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("dddcc0710eb6cf9efd5b920a8453e1e07157bddd"),
            next_position_offset=0,
            successive_visits=3,
        ),
    )


def test_journal_client_origin_visit_status_from_journal_last_uneventful(swh_scheduler):
    visit_status = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE3 + ONE_DAY,
        "type": "git",
        "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
    }

    # Let's insert some visit stats with some previous visit information
    swh_scheduler.origin_visit_stats_upsert(
        [
            OriginVisitStats(
                url=visit_status["origin"],
                visit_type=visit_status["type"],
                last_successful=DATE2,
                last_visit=DATE3,
                last_visit_status=LastVisitStatus.failed,
                last_snapshot=visit_status["snapshot"],
                next_visit_queue_position=None,
                next_position_offset=4,
                successive_visits=1,
            )
        ]
    )

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get(
        [(visit_status["origin"], visit_status["type"])]
    )

    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url=visit_status["origin"],
            visit_type=visit_status["type"],
            last_visit=DATE3 + ONE_DAY,
            last_successful=DATE3 + ONE_DAY,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=visit_status["snapshot"],
            next_visit_queue_position=None,
            next_position_offset=5,
            successive_visits=1,
        ),
    )


VISIT_STATUSES = [
    {**ovs, "date": DATE1 + n * ONE_DAY}
    for n, ovs in enumerate(
        [
            {
                "origin": "foo",
                "type": "git",
                "visit": 1,
                "status": "created",
                "snapshot": None,
            },
            {
                "origin": "foo",
                "type": "git",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "foo",
                "type": "git",
                "visit": 2,
                "status": "created",
                "snapshot": None,
            },
            {
                "origin": "foo",
                "type": "git",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
        ]
    )
]


@pytest.mark.parametrize(
    "visit_statuses", permutations(VISIT_STATUSES, len(VISIT_STATUSES))
)
def test_journal_client_origin_visit_status_permutation0(visit_statuses, swh_scheduler):
    """Ensure out of order topic subscription ends up in the same final state"""
    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    visit_stats = actual_origin_visit_stats[0]
    assert_visit_stats_ok(
        visit_stats,
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE1 + 3 * ONE_DAY,
            last_visit=DATE1 + 3 * ONE_DAY,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        ),
        ignore_fields=[
            "next_visit_queue_position",
            "next_position_offset",
            "successive_visits",
        ],
    )

    # We ignore out of order messages, so the next_position_offset isn't exact
    # depending on the permutation. What matters is consistency of the final
    # dates (last_visit and last_successful).
    assert 4 <= visit_stats.next_position_offset <= 5
    # same goes for successive_visits
    assert 1 <= visit_stats.successive_visits <= 2


VISIT_STATUSES_1 = [
    {**ovs, "date": DATE1 + n * ONE_DAY}
    for n, ovs in enumerate(
        [
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 1,
                "status": "partial",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 3,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 4,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
        ]
    )
]


@pytest.mark.parametrize(
    "visit_statuses", permutations(VISIT_STATUSES_1, len(VISIT_STATUSES_1))
)
def test_journal_client_origin_visit_status_permutation1(visit_statuses, swh_scheduler):
    """Ensure out of order topic subscription ends up in the same final state"""
    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_visit_stats = swh_scheduler.origin_visit_stats_get([("cavabarder", "hg")])
    visit_stats = actual_visit_stats[0]

    assert_visit_stats_ok(
        visit_stats,
        OriginVisitStats(
            url="cavabarder",
            visit_type="hg",
            last_successful=DATE1 + 3 * ONE_DAY,
            last_visit=DATE1 + 3 * ONE_DAY,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
        ),
        ignore_fields=[
            "next_visit_queue_position",
            "next_position_offset",
            "successive_visits",
        ],
    )

    # We ignore out of order messages, so the next_position_offset isn't exact
    # depending on the permutation. What matters is consistency of the final
    # dates (last_visit and last_successful).
    assert 2 <= visit_stats.next_position_offset <= 5
    # same goes for successive_visits
    assert 1 <= visit_stats.successive_visits <= 4


VISIT_STATUSES_2 = [
    {**ovs, "date": DATE1 + n * ONE_DAY}
    for n, ovs in enumerate(
        [
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("0000000000000000000000000000000000000000"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("1111111111111111111111111111111111111111"),
            },
            {
                "origin": "iciaussi",
                "type": "hg",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("2222222222222222222222222222222222222222"),
            },
            {
                "origin": "iciaussi",
                "type": "hg",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("3333333333333333333333333333333333333333"),
            },
            {
                "origin": "cavabarder",
                "type": "git",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("4444444444444444444444444444444444444444"),
            },
            {
                "origin": "cavabarder",
                "type": "git",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("5555555555555555555555555555555555555555"),
            },
            {
                "origin": "iciaussi",
                "type": "git",
                "visit": 1,
                "status": "full",
                "snapshot": hash_to_bytes("6666666666666666666666666666666666666666"),
            },
            {
                "origin": "iciaussi",
                "type": "git",
                "visit": 2,
                "status": "full",
                "snapshot": hash_to_bytes("7777777777777777777777777777777777777777"),
            },
        ]
    )
]


def test_journal_client_origin_visit_status_after_grab_next_visits(
    swh_scheduler, stored_lister
):
    """Ensure OriginVisitStat entries created in the db as a result of calling
    grab_next_visits() do not mess the OriginVisitStats upsert mechanism.

    """

    listed_origins = [
        ListedOrigin(lister_id=stored_lister.id, url=url, visit_type=visit_type)
        for (url, visit_type) in set((v["origin"], v["type"]) for v in VISIT_STATUSES_2)
    ]
    swh_scheduler.record_listed_origins(listed_origins)
    before = utcnow()
    swh_scheduler.grab_next_visits(
        visit_type="git", count=10, policy="oldest_scheduled_first"
    )
    after = utcnow()

    assert swh_scheduler.origin_visit_stats_get([("cavabarder", "hg")]) == []
    assert swh_scheduler.origin_visit_stats_get([("cavabarder", "git")])[0] is not None

    process_journal_objects(
        {"origin_visit_status": VISIT_STATUSES_2}, scheduler=swh_scheduler
    )

    for url in ("cavabarder", "iciaussi"):
        ovs = swh_scheduler.origin_visit_stats_get([(url, "git")])[0]
        assert before <= ovs.last_scheduled <= after

        ovs = swh_scheduler.origin_visit_stats_get([(url, "hg")])[0]
        assert ovs.last_scheduled is None

    ovs = swh_scheduler.origin_visit_stats_get([("cavabarder", "git")])[0]
    assert ovs.last_successful == DATE1 + 5 * ONE_DAY
    assert ovs.last_visit == DATE1 + 5 * ONE_DAY
    assert ovs.last_visit_status == LastVisitStatus.successful
    assert ovs.last_snapshot == hash_to_bytes(
        "5555555555555555555555555555555555555555"
    )


def test_journal_client_origin_visit_status_duplicated_messages(swh_scheduler):
    """A duplicated message must be ignored"""
    visit_status = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE1,
        "type": "git",
        "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
    }

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    process_journal_objects(
        {"origin_visit_status": [visit_status]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE1,
            last_visit=DATE1,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            successive_visits=1,
        ),
    )


def test_journal_client_origin_visit_status_several_upsert(swh_scheduler):
    """An old message updates old information"""
    visit_status1 = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE1,
        "type": "git",
        "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
    }

    visit_status2 = {
        "origin": "foo",
        "visit": 1,
        "status": "full",
        "date": DATE2,
        "type": "git",
        "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
    }

    process_journal_objects(
        {"origin_visit_status": [visit_status2]}, scheduler=swh_scheduler
    )

    process_journal_objects(
        {"origin_visit_status": [visit_status1]}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("foo", "git")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="foo",
            visit_type="git",
            last_successful=DATE2,
            last_visit=DATE2,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            next_position_offset=4,
            successive_visits=1,
        ),
    )


VISIT_STATUSES_SAME_SNAPSHOT = [
    {**ovs, "date": DATE1 + n * ONE_YEAR}
    for n, ovs in enumerate(
        [
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 3,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 4,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
            {
                "origin": "cavabarder",
                "type": "hg",
                "visit": 4,
                "status": "full",
                "snapshot": hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
            },
        ]
    )
]


@pytest.mark.parametrize(
    "visit_statuses",
    permutations(VISIT_STATUSES_SAME_SNAPSHOT, len(VISIT_STATUSES_SAME_SNAPSHOT)),
)
def test_journal_client_origin_visit_statuses_same_snapshot_permutation(
    visit_statuses, swh_scheduler
):
    """Ensure out of order topic subscription ends up in the same final state"""
    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get(
        [("cavabarder", "hg")]
    )
    visit_stats = actual_origin_visit_stats[0]
    assert_visit_stats_ok(
        visit_stats,
        OriginVisitStats(
            url="cavabarder",
            visit_type="hg",
            last_successful=DATE1 + 2 * ONE_YEAR,
            last_visit=DATE1 + 2 * ONE_YEAR,
            last_visit_status=LastVisitStatus.successful,
            last_snapshot=hash_to_bytes("aaaaaabbbeb6cf9efd5b920a8453e1e07157b6cd"),
        ),
        ignore_fields=[
            "next_visit_queue_position",
            "next_position_offset",
            "successive_visits",
        ],
    )

    # We ignore out of order messages, so the next_position_offset isn't exact
    # depending on the permutation. What matters is consistency of the final
    # dates (last_visit and last_successful).
    assert 4 <= visit_stats.next_position_offset <= 6
    # same goes for successive_visits
    assert 1 <= visit_stats.successive_visits <= 3


@pytest.mark.parametrize(
    "position_offset, interval",
    [
        (0, 1),
        (1, 1),
        (2, 2),
        (3, 2),
        (4, 2),
        (5, 4),
        (6, 16),
        (7, 64),
        (8, 256),
        (9, 1024),
        (10, 4096),
        (11, 16384),
    ],
)
def test_journal_client_from_position_offset_to_days(position_offset, interval):
    assert from_position_offset_to_days(position_offset) == interval


def test_journal_client_from_position_offset_to_days_only_positive_input():
    with pytest.raises(AssertionError):
        from_position_offset_to_days(-1)


@pytest.mark.parametrize(
    "fudge_factor,next_position_offset",
    [
        (0.01, 1),
        (-0.01, 5),
        (0.1, 8),
        (-0.1, 10),
    ],
)
def test_next_visit_queue_position(mocker, fudge_factor, next_position_offset):
    mock_random = mocker.patch("swh.scheduler.journal_client.random.uniform")
    mock_random.return_value = fudge_factor

    actual_position = next_visit_queue_position(
        {},
        {
            "next_position_offset": next_position_offset,
            "visit_type": "svn",
        },
    )

    assert actual_position == int(
        24
        * 3600
        * from_position_offset_to_days(next_position_offset)
        * (1 + fudge_factor)
    )

    assert mock_random.called


@pytest.mark.parametrize(
    "fudge_factor,next_position_offset",
    [
        (0.02, 2),
        (-0.02, 3),
        (0, 7),
        (-0.09, 9),
    ],
)
def test_next_visit_queue_position_with_state(
    mocker, fudge_factor, next_position_offset
):
    mock_random = mocker.patch("swh.scheduler.journal_client.random.uniform")
    mock_random.return_value = fudge_factor

    actual_position = next_visit_queue_position(
        {"git": 0},
        {
            "next_position_offset": next_position_offset,
            "visit_type": "git",
        },
    )

    assert actual_position == int(
        24
        * 3600
        * from_position_offset_to_days(next_position_offset)
        * (1 + fudge_factor)
    )

    assert mock_random.called


@pytest.mark.parametrize(
    "fudge_factor,next_position_offset",
    [
        (0.03, 3),
        (-0.03, 4),
        (0.08, 7),
        (-0.08, 9),
    ],
)
def test_next_visit_queue_position_with_next_visit_queue(
    mocker, fudge_factor, next_position_offset
):
    mock_random = mocker.patch("swh.scheduler.journal_client.random.uniform")
    mock_random.return_value = fudge_factor

    actual_position = next_visit_queue_position(
        {},
        {
            "next_position_offset": next_position_offset,
            "visit_type": "hg",
            "next_visit_queue_position": 0,
        },
    )

    assert actual_position == int(
        24
        * 3600
        * from_position_offset_to_days(next_position_offset)
        * (1 + fudge_factor)
    )

    assert mock_random.called


def test_disable_failing_origins(swh_scheduler):
    """Origin with too many failed attempts ends up being deactivated in the scheduler."""

    # actually store the origin in the scheduler so we can check it's deactivated in the
    # end.
    lister = swh_scheduler.get_or_create_lister(
        name="something", instance_name="something"
    )
    origin = ListedOrigin(
        url="bar", enabled=True, visit_type="svn", lister_id=lister.id
    )
    swh_scheduler.record_listed_origins([origin])

    visit_statuses = [
        {
            "origin": "bar",
            "visit": 2,
            "status": "failed",
            "date": DATE1,
            "type": "svn",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "failed",
            "date": DATE2,
            "type": "svn",
            "snapshot": None,
        },
        {
            "origin": "bar",
            "visit": 3,
            "status": "failed",
            "date": DATE3,
            "type": "svn",
            "snapshot": None,
        },
    ]

    process_journal_objects(
        {"origin_visit_status": visit_statuses}, scheduler=swh_scheduler
    )

    actual_origin_visit_stats = swh_scheduler.origin_visit_stats_get([("bar", "svn")])
    assert_visit_stats_ok(
        actual_origin_visit_stats[0],
        OriginVisitStats(
            url="bar",
            visit_type="svn",
            last_successful=None,
            last_visit=DATE3,
            last_visit_status=LastVisitStatus.failed,
            next_position_offset=6,
            successive_visits=3,
        ),
    )

    # Now check that the origin in question is disabled
    actual_page = swh_scheduler.get_listed_origins(url="bar", enabled=False)

    assert len(actual_page.results) == 1
    assert actual_page.next_page_token is None

    for origin in actual_page.results:
        assert origin.enabled is False
        assert origin.lister_id == lister.id
        assert origin.url == "bar"
        assert origin.visit_type == "svn"



# Copyright (C) 2019-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Module in charge of testing the scheduler runner module"""

from itertools import count
from time import sleep

from celery.result import AsyncResult, GroupResult
from kombu import Exchange, Queue
import pytest

from swh.scheduler.celery_backend.runner import run_ready_tasks
from swh.scheduler.tests.tasks import (
    TASK_ADD,
    TASK_ECHO,
    TASK_ERROR,
    TASK_MULTIPING,
    TASK_PING,
)
from swh.scheduler.utils import create_task_dict

# Queues to subscribe. Due to the rerouting of high priority tasks, this module requires
# to declare all queues/task names
TEST_QUEUES = [
    "celery",
    TASK_ECHO,
    TASK_ERROR,
    TASK_PING,
    TASK_ADD,
    TASK_MULTIPING,
    # and the high priority queue
    f"save_code_now:{TASK_ADD}",
]


@pytest.fixture(scope="session")
def swh_scheduler_celery_app(swh_scheduler_celery_app):
    swh_scheduler_celery_app.add_defaults(
        {
            "task_queues": [
                Queue(queue, Exchange(queue), routing_key=queue)
                for queue in TEST_QUEUES
            ],
        }
    )
    return swh_scheduler_celery_app


def test_ping(swh_scheduler_celery_app, swh_scheduler_celery_worker):
    res = swh_scheduler_celery_app.send_task(TASK_PING)
    assert res
    res.wait()
    assert res.successful()
    assert res.result == "OK"


def test_ping_with_kw(swh_scheduler_celery_app, swh_scheduler_celery_worker):
    res = swh_scheduler_celery_app.send_task(TASK_PING, kwargs={"a": 1})
    assert res
    res.wait()
    assert res.successful()
    assert res.result == "OK (kw={'a': 1})"


def test_multiping(swh_scheduler_celery_app, swh_scheduler_celery_worker):
    "Test that a task that spawns subtasks (group) works"
    res = swh_scheduler_celery_app.send_task(TASK_MULTIPING, kwargs={"n": 5})
    assert res

    res.wait()
    assert res.successful()

    # retrieve the GroupResult for this task and wait for all the subtasks
    # to complete
    promise_id = res.result
    assert promise_id
    promise = GroupResult.restore(promise_id, app=swh_scheduler_celery_app)
    for i in range(5):
        if promise.ready():
            break
        sleep(1)

    results = [x.get() for x in promise.results]
    assert len(results) == 5
    for i in range(5):
        assert ("OK (kw={'i': %s})" % i) in results


def test_scheduler_fixture(
    swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_scheduler
):
    "Test that the scheduler fixture works properly"
    task_type = swh_scheduler.get_task_type("swh-test-ping")

    assert task_type
    assert task_type["backend_name"] == TASK_PING

    swh_scheduler.create_tasks([create_task_dict("swh-test-ping", "oneshot")])

    backend_tasks = run_ready_tasks(swh_scheduler, swh_scheduler_celery_app)
    assert backend_tasks
    for task in backend_tasks:
        # Make sure the task completed
        AsyncResult(id=task["backend_id"]).get()


def test_run_ready_task_standard(
    swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_scheduler
):
    """Ensure scheduler runner schedules tasks ready for scheduling"""
    task_type_name, backend_name = "swh-test-add", TASK_ADD
    task_type = swh_scheduler.get_task_type(task_type_name)
    assert task_type
    assert task_type["backend_name"] == backend_name

    task_inputs = [
        ("oneshot", (12, 30)),
        ("oneshot", (20, 10)),
        ("recurring", (30, 10)),
    ]

    tasks = swh_scheduler.create_tasks(
        create_task_dict(task_type_name, policy, *args)
        for (policy, args) in task_inputs
    )

    assert len(tasks) == len(task_inputs)

    task_ids = set()
    for task in tasks:
        assert task["status"] == "next_run_not_scheduled"
        assert task["priority"] is None
        task_ids.add(task["id"])

    backend_tasks = run_ready_tasks(swh_scheduler, swh_scheduler_celery_app)
    assert len(backend_tasks) == len(tasks)

    scheduled_tasks = swh_scheduler.search_tasks(task_type=task_type_name)
    assert len(scheduled_tasks) == len(tasks)
    for task in scheduled_tasks:
        assert task["status"] == "next_run_scheduled"
        assert task["id"] in task_ids

    # Ensure each task is indeed scheduled to the queue backend
    for i, (_, args) in enumerate(task_inputs):
        task = backend_tasks[i]
        value = AsyncResult(id=task["backend_id"]).get()
        assert value == sum(args)


def test_run_ready_task_with_priority(
    swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_scheduler
):
    """Ensure scheduler runner schedules priority tasks ready for scheduling"""
    task_type_name, backend_name = "swh-test-add", TASK_ADD
    task_type = swh_scheduler.get_task_type(task_type_name)
    assert task_type
    assert task_type["backend_name"] == backend_name

    task_inputs = [
        ("oneshot", (10, 22), "low"),
        ("oneshot", (20, 10), "normal"),
        ("recurring", (30, 10), "high"),
    ]

    tasks = swh_scheduler.create_tasks(
        create_task_dict(task_type_name, policy, *args, priority=priority)
        for (policy, args, priority) in task_inputs
    )

    assert len(tasks) == len(task_inputs)

    task_ids = set()
    for task in tasks:
        assert task["status"] == "next_run_not_scheduled"
        assert task["priority"] is not None
        task_ids.add(task["id"])

    backend_tasks = run_ready_tasks(
        swh_scheduler, swh_scheduler_celery_app, task_types=[], with_priority=True
    )
    assert len(backend_tasks) == len(tasks)

    scheduled_tasks = swh_scheduler.search_tasks(task_type=task_type_name)
    assert len(scheduled_tasks) == len(tasks)
    for task in scheduled_tasks:
        assert task["status"] == "next_run_scheduled"
        assert task["id"] in task_ids

    # Ensure each priority task is indeed scheduled to the queue backend
    for i, (_, args, _) in enumerate(task_inputs):
        task = backend_tasks[i]
        value = AsyncResult(id=task["backend_id"]).get()
        assert value == sum(args)


def test_task_exception(
    swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_scheduler
):
    task_type = swh_scheduler.get_task_type("swh-test-error")
    assert task_type
    assert task_type["backend_name"] == TASK_ERROR

    swh_scheduler.create_tasks([create_task_dict("swh-test-error", "oneshot")])

    backend_tasks = run_ready_tasks(swh_scheduler, swh_scheduler_celery_app)
    assert len(backend_tasks) == 1

    task = backend_tasks[0]
    result = AsyncResult(id=task["backend_id"])
    with pytest.raises(NotImplementedError):
        result.get()


def test_statsd(swh_scheduler_celery_app, swh_scheduler_celery_worker, mocker):
    m = mocker.patch("swh.scheduler.task.Statsd._send_to_server")
    mocker.patch("swh.scheduler.task.ts", side_effect=count())
    mocker.patch("swh.core.statsd.monotonic", side_effect=count())
    res = swh_scheduler_celery_app.send_task(TASK_ECHO)
    assert res
    res.wait()
    assert res.successful()
    assert res.result == {}

    m.assert_any_call(
        "swh_task_called_count:1|c|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_start_ts:0|g|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_end_ts:1|g|"
        "#status:uneventful,task:swh.scheduler.tests.tasks.echo,"
        "worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_duration_seconds:1000|ms|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_success_count:1|c|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )


def test_statsd_with_status(
    swh_scheduler_celery_app, swh_scheduler_celery_worker, mocker
):
    m = mocker.patch("swh.scheduler.task.Statsd._send_to_server")
    mocker.patch("swh.scheduler.task.ts", side_effect=count())
    mocker.patch("swh.core.statsd.monotonic", side_effect=count())
    res = swh_scheduler_celery_app.send_task(TASK_ECHO, kwargs={"status": "eventful"})
    assert res
    res.wait()
    assert res.successful()
    assert res.result == {"status": "eventful"}

    m.assert_any_call(
        "swh_task_called_count:1|c|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_start_ts:0|g|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_end_ts:1|g|"
        "#status:eventful,task:swh.scheduler.tests.tasks.echo,"
        "worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_duration_seconds:1000|ms|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )
    m.assert_any_call(
        "swh_task_success_count:1|c|"
        "#task:swh.scheduler.tests.tasks.echo,worker:unknown worker"
    )



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta
import logging
from queue import Queue
import time

import pytest
import yaml

from swh.scheduler.celery_backend.recurrent_visits import (
    DEFAULT_DVCS_POLICY,
    VisitSchedulerThreads,
    grab_next_visits_policy_weights,
    send_visits_for_visit_type,
    spawn_visit_scheduler_thread,
    terminate_visit_scheduler_threads,
    visit_scheduler_thread,
)

from .test_cli import invoke

TEST_MAX_QUEUE = 10000
MODULE_NAME = "swh.scheduler.celery_backend.recurrent_visits"


def _compute_backend_name(visit_type: str) -> str:
    "Build a dummy reproducible backend name"
    return f"swh.loader.{visit_type}.tasks"


@pytest.fixture
def swh_scheduler(swh_scheduler):
    """Override default fixture of the scheduler to install some more task types."""
    for visit_type in ["test-git", "test-hg", "test-svn"]:
        task_type = f"load-{visit_type}"
        swh_scheduler.create_task_type(
            {
                "type": task_type,
                "max_queue_length": TEST_MAX_QUEUE,
                "description": "The {} testing task".format(task_type),
                "backend_name": _compute_backend_name(visit_type),
                "default_interval": timedelta(days=1),
                "min_interval": timedelta(hours=6),
                "max_interval": timedelta(days=12),
            }
        )
    return swh_scheduler


@pytest.fixture
def all_task_types(swh_scheduler):
    return {
        task_type_d["type"]: task_type_d
        for task_type_d in swh_scheduler.get_task_types()
    }


def test_cli_schedule_recurrent_unknown_visit_type(swh_scheduler):
    """When passed an unknown visit type, the recurrent visit scheduler should refuse
    to start."""

    with pytest.raises(ValueError, match="Unknown"):
        invoke(
            swh_scheduler,
            False,
            [
                "schedule-recurrent",
                "--visit-type",
                "unknown",
                "--visit-type",
                "test-git",
            ],
        )


def test_cli_schedule_recurrent_noop(swh_scheduler, mocker):
    """When passing no visit types, the recurrent visit scheduler should start."""

    spawn_visit_scheduler_thread = mocker.patch(
        f"{MODULE_NAME}.spawn_visit_scheduler_thread"
    )
    spawn_visit_scheduler_thread.side_effect = SystemExit
    # The actual scheduling threads won't spawn, they'll immediately terminate. This
    # only exercises the logic to pull task types out of the database

    result = invoke(swh_scheduler, False, ["schedule-recurrent"])
    assert result.exit_code == 0, result.output


def test_send_visits_for_type_no_origin_scheduled_backoff(
    swh_scheduler, all_task_types, mocker
):
    visit_type = "test-git"
    task_type = f"load-{visit_type}"
    mocker.patch.object(time, "monotonic", lambda: 0)
    no_origins_scheduled_backoff = 60
    assert (
        send_visits_for_visit_type(
            swh_scheduler,
            mocker.MagicMock(),
            visit_type,
            all_task_types[task_type],
            DEFAULT_DVCS_POLICY,
            no_origins_scheduled_backoff,
        )
        == no_origins_scheduled_backoff
    )


def test_cli_schedule_recurrent_no_origins_scheduled_backoff_in_config(
    swh_scheduler, mocker
):
    """When passing no visit types, the recurrent visit scheduler should start."""

    config = """
    scheduler:
        cls: foo
        args: {}
    no_origins_scheduled_backoff: 60
    """

    spawn_visit_scheduler_thread = mocker.patch(
        f"{MODULE_NAME}.spawn_visit_scheduler_thread"
    )
    spawn_visit_scheduler_thread.side_effect = SystemExit
    # The actual scheduling threads won't spawn, they'll immediately terminate. This
    # only exercises the logic to pull task types out of the database

    result = invoke(
        swh_scheduler,
        False,
        ["schedule-recurrent", "--visit-type", "test-git"],
        config=config,
    )
    assert result.exit_code == 0, result.output

    assert yaml.safe_load(config) in spawn_visit_scheduler_thread.call_args[0]


def test_recurrent_visit_scheduling(
    swh_scheduler,
    caplog,
    listed_origins_by_type,
    all_task_types,
    mocker,
):
    """Scheduling known tasks is ok."""

    caplog.set_level(logging.DEBUG, MODULE_NAME)
    nb_origins = 1000

    mock_celery_app = mocker.MagicMock()
    mock_available_slots = mocker.patch(f"{MODULE_NAME}.get_available_slots")
    mock_available_slots.return_value = nb_origins  # Slots available in queue

    # Make sure the scheduler is properly configured in terms of visit/task types
    visit_types = list(listed_origins_by_type.keys())
    assert len(visit_types) > 0

    task_types = []
    origins = []
    for visit_type, _origins in listed_origins_by_type.items():
        origins.extend(swh_scheduler.record_listed_origins(_origins))
        task_type_name = f"load-{visit_type}"
        assert task_type_name in all_task_types.keys()
        task_type = all_task_types[task_type_name]
        task_type["visit_type"] = visit_type
        # we'll limit the orchestrator to the origins' type we know
        task_types.append(task_type)

    for visit_type in ["test-git", "test-svn"]:
        task_type = f"load-{visit_type}"
        send_visits_for_visit_type(
            swh_scheduler,
            mock_celery_app,
            visit_type,
            all_task_types[task_type],
            DEFAULT_DVCS_POLICY,
        )

    assert mock_available_slots.called, "The available slots functions should be called"

    records = [record.message for record in caplog.records]

    # Mapping over the dict ratio/policies entries can change overall order so let's
    # check the set of records
    expected_records = set()
    for task_type in task_types:
        visit_type = task_type["visit_type"]
        queue_name = task_type["backend_name"]
        msg = (
            f"{nb_origins} available slots for visit type {visit_type} "
            f"in queue {queue_name}"
        )
        expected_records.add(msg)

    for expected_record in expected_records:
        assert expected_record in set(records)


@pytest.mark.parametrize(
    "visit_type, extras",
    [("test-hg", {}), ("test-git", {"tablesample": 0.1})],
)
def test_recurrent_visit_additional_parameters(
    swh_scheduler, mocker, visit_type, extras
):
    """Testing additional policy parameters"""

    mock_grab_next_visits = mocker.patch.object(swh_scheduler, "grab_next_visits")
    mock_grab_next_visits.return_value = []
    policy_cfg = DEFAULT_DVCS_POLICY[:]
    for policy in policy_cfg:
        policy.update(extras)
    grab_next_visits_policy_weights(swh_scheduler, visit_type, 10, policy_cfg)

    for call in mock_grab_next_visits.call_args_list:
        assert call[1].get("tablesample") == extras.get("tablesample")


@pytest.fixture
def scheduler_config(swh_scheduler_config):
    return {"scheduler": {"cls": "postgresql", **swh_scheduler_config}, "celery": {}}


def test_visit_scheduler_thread_unknown_task(
    swh_scheduler,
    scheduler_config,
):
    """Starting a thread with unknown task type reports the error"""

    unknown_visit_type = "unknown"
    command_queue = Queue()
    exc_queue = Queue()

    visit_scheduler_thread(
        scheduler_config, unknown_visit_type, command_queue, exc_queue
    )

    assert command_queue.empty() is True
    assert exc_queue.empty() is False
    assert len(exc_queue.queue) == 1
    result = exc_queue.queue.pop()
    assert result[0] == unknown_visit_type
    assert isinstance(result[1], ValueError)


def test_spawn_visit_scheduler_thread_noop(scheduler_config, visit_types, mocker):
    """Spawning and terminating threads runs smoothly"""

    threads: VisitSchedulerThreads = {}
    exc_queue = Queue()
    mock_build_app = mocker.patch("swh.scheduler.celery_backend.config.build_app")
    mock_build_app.return_value = mocker.MagicMock()

    assert len(threads) == 0
    for visit_type in visit_types:
        spawn_visit_scheduler_thread(threads, exc_queue, scheduler_config, visit_type)

    # This actually only checks the spawning and terminating logic is sound

    assert len(threads) == len(visit_types)

    actual_threads = terminate_visit_scheduler_threads(threads)
    assert not len(actual_threads)

    assert mock_build_app.called



# Copyright (C) 2017-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import datetime
from typing import Dict, List, Optional

TEMPLATES = {
    "test-git": {
        "type": "load-test-git",
        "arguments": {
            "args": [],
            "kwargs": {},
        },
        "next_run": None,
    },
    "test-hg": {
        "type": "load-test-hg",
        "arguments": {
            "args": [],
            "kwargs": {},
        },
        "next_run": None,
        "policy": "oneshot",
    },
}


TASK_TYPES = {
    "test-git": {
        "type": "load-test-git",
        "description": "Update a git repository",
        "backend_name": "swh.loader.git.tasks.UpdateGitRepository",
        "default_interval": datetime.timedelta(days=64),
        "min_interval": datetime.timedelta(hours=12),
        "max_interval": datetime.timedelta(days=64),
        "backoff_factor": 2,
        "max_queue_length": None,
        "num_retries": 7,
        "retry_delay": datetime.timedelta(hours=2),
    },
    "test-hg": {
        "type": "load-test-hg",
        "description": "Update a mercurial repository",
        "backend_name": "swh.loader.mercurial.tasks.UpdateHgRepository",
        "default_interval": datetime.timedelta(days=64),
        "min_interval": datetime.timedelta(hours=12),
        "max_interval": datetime.timedelta(days=64),
        "backoff_factor": 2,
        "max_queue_length": None,
        "num_retries": 7,
        "retry_delay": datetime.timedelta(hours=2),
    },
}


def _task_from_template(
    template: Dict,
    next_run: datetime.datetime,
    priority: Optional[str],
    *args,
    **kwargs,
) -> Dict:
    ret = copy.deepcopy(template)
    ret["next_run"] = next_run
    if priority:
        ret["priority"] = priority
    if args:
        ret["arguments"]["args"] = list(args)
    if kwargs:
        ret["arguments"]["kwargs"] = kwargs
    return ret


def tasks_from_template(
    template: Dict,
    max_timestamp: datetime.datetime,
    num: Optional[int] = None,
    priority: Optional[str] = None,
    num_priorities: Dict[Optional[str], int] = {},
) -> List[Dict]:
    """Build ``num`` tasks from template"""
    assert bool(num) != bool(num_priorities), "mutually exclusive"
    if not num_priorities:
        assert num is not None  # to please mypy
        num_priorities = {None: num}
    tasks: List[Dict] = []
    for (priority, num) in num_priorities.items():
        for _ in range(num):
            i = len(tasks)
            tasks.append(
                _task_from_template(
                    template,
                    max_timestamp - datetime.timedelta(microseconds=i),
                    priority,
                    "argument-%03d" % i,
                    **{"kwarg%03d" % i: "bogus-kwarg"},
                )
            )
    return tasks


def tasks_with_priority_from_template(
    template: Dict, max_timestamp: datetime.datetime, num: int, priority: str
) -> List[Dict]:
    """Build tasks with priority from template"""
    return [
        _task_from_template(
            template,
            max_timestamp - datetime.timedelta(microseconds=i),
            priority,
            "argument-%03d" % i,
            **{"kwarg%03d" % i: "bogus-kwarg"},
        )
        for i in range(num)
    ]


LISTERS = (
    {"name": "github", "instance_name": "github"},
    {"name": "gitlab", "instance_name": "gitlab"},
    {"name": "gitlab", "instance_name": "freedesktop"},
    {"name": "npm"},
    {"name": "pypi"},
)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from hypothesis import settings

# define tests profile. Full documentation is at:
# https://hypothesis.readthedocs.io/en/latest/settings.html#settings-profiles
settings.register_profile("fast", max_examples=5, deadline=5000)
settings.register_profile("slow", max_examples=20, deadline=5000)


pytest_plugins = [
    "swh.scheduler.pytest_plugin",
    "swh.core.db.pytest_plugin",
    "swh.storage.pytest_plugin",
]



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import cli

if TYPE_CHECKING:
    from typing import Dict, List, Optional


@cli.group("add-forge-now")
@click.option(
    "-p",
    "--preset",
    "preset",
    default="production",
    type=click.Choice(["production", "staging"]),
    help='Determine preset to use, "production" by default.',
)
@click.pass_context
def add_forge_now(ctx, preset):
    """Manipulate add-forge-now requests."""
    if not ctx.obj["scheduler"]:
        raise ValueError("Scheduler class (local/remote) must be instantiated")

    ctx.obj["preset"] = preset


@add_forge_now.command("register-lister")
@click.argument("lister_name", nargs=1, required=True)
@click.argument("options", nargs=-1)
@click.pass_context
def register_lister_cli(
    ctx,
    lister_name,
    options,
):
    """Register the lister tasks in the scheduler.

    The specifics of what tasks are registered depends on the add-forge-now --preset
    option:
    - staging preset: a single oneshot full listing task is scheduled. This "full"
      listing is limited to 3 pages and 10 origins per page. The origins are recorded as
      disabled (to avoid their recurrent loading).
    - production preset: a recurrent full and incremental (if the loader has such a
      task) listing task are scheduled. The first run of the full lister is scheduled
      immediately, and the first run of the incremental lister is delayed by a day.

    """
    from .utils import lister_task_type, parse_options, task_add

    scheduler = ctx.obj["scheduler"]
    preset = ctx.obj["preset"]

    # Map the associated task types for the lister
    task_type_names: Dict[str, str] = {
        listing_type: lister_task_type(lister_name, listing_type)
        for listing_type in ["full", "incremental", None]
    }

    task_types: Dict[str, Dict] = {}
    for listing_type, task_type_name in task_type_names.items():
        task_type = scheduler.get_task_type(task_type_name)
        if task_type:
            task_types[listing_type] = task_type

    if not task_types:
        raise ValueError(f"Unknown lister type {lister_name}.")

    (args, kw) = parse_options(options)

    # Recurring policy on production
    if preset == "production":
        policy = "recurring"
    else:  # staging, "full" but limited listing as a oneshot
        policy = "oneshot"
        kw.update({"max_pages": 3, "max_origins_per_page": 10, "enable_origins": False})
        # We want a "full" listing in production if both incremental and full exists
        if "full" in task_types:
            task_types.pop("incremental", None)

    from datetime import timedelta

    from swh.scheduler.utils import utcnow

    for listing_type, task_type in task_types.items():
        now = utcnow()
        next_run = now if listing_type == "full" else now + timedelta(days=1)
        task_add(
            scheduler,
            task_type_name=task_type["type"],
            args=args,
            kw=kw,
            policy=policy,
            next_run=next_run,
        )


@add_forge_now.command("schedule-first-visits")
@click.option(
    "--type-name",
    "-t",
    "visit_type_names",
    help="Visit/loader type (can be provided multiple times)",
    type=str,
    multiple=True,
)
@click.option(
    "--lister-name",
    default=None,
    help="Limit origins to those listed from lister with provided name",
)
@click.option(
    "--lister-instance-name",
    default=None,
    help="Limit origins to those listed from lister with instance name",
)
@click.pass_context
def schedule_first_visits_cli(
    ctx,
    visit_type_names: List[str],
    lister_name: Optional[str] = None,
    lister_instance_name: Optional[str] = None,
):
    """Send next origin visits of VISIT_TYPE_NAME(S) loader to celery, filling the
    associated add_forge_now queue(s).

    """
    from .utils import get_task_type, send_to_celery

    scheduler = ctx.obj["scheduler"]
    preset = ctx.obj["preset"]

    visit_type_to_queue: Dict[str, str] = {}
    unknown_task_types = []
    for visit_type_name in visit_type_names:
        task_type = get_task_type(scheduler, visit_type_name)
        if not task_type:
            unknown_task_types.append(visit_type_name)
            continue
        queue_name = task_type["backend_name"]
        visit_type_to_queue[visit_type_name] = f"add_forge_now:{queue_name}"

    if unknown_task_types:
        raise ValueError(f"Unknown task types {','.join(unknown_task_types)}.")

    send_to_celery(
        scheduler,
        visit_type_to_queue=visit_type_to_queue,
        enabled=preset == "production",
        lister_name=lister_name,
        lister_instance_name=lister_instance_name,
    )
